# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '5.6.1'
git_version = '490762ca26f15c966a8b35a322ba91b27515c877'
full_version = '5.6.1'
is_release = True

if not is_release:
    version = full_version